@@ -13,10 +13,14 @@ const DERIVE_NAME: &str = "Accounts";
 const ERROR_CODE_OFFSET: u32 = 300;
 
 // Parse an entire interface file.
-pub fn parse(filename: impl AsRef<Path>) -> Result<Idl> {
+pub fn parse(filename: impl AsRef<Path>) -> Result<Option<Idl>> {
     let ctx = CrateContext::parse(filename)?;
 
-    let p = program::parse(parse_program_mod(&ctx))?;
+    let program_mod = match parse_program_mod(&ctx) {
+        None => return Ok(None),
+        Some(m) => m,
+    };
+    let p = program::parse(program_mod)?;
 
     let accs = parse_account_derives(&ctx);
 
@@ -218,7 +222,7 @@ pub fn parse(filename: impl AsRef<Path>) -> Result<Idl> {
         }
     }
 
-    Ok(Idl {
+    Ok(Some(Idl {
         version: "0.0.0".to_string(),
         name: p.name.to_string(),
         state,
@@ -232,11 +236,11 @@ pub fn parse(filename: impl AsRef<Path>) -> Result<Idl> {
         },
         errors: error_codes,
         metadata: None,
-    })
+    }))
 }
 
 // Parse the main program mod.
-fn parse_program_mod(ctx: &CrateContext) -> syn::ItemMod {
+fn parse_program_mod(ctx: &CrateContext) -> Option<syn::ItemMod> {
     let root = ctx.root_module();
     let mods = root
         .items()
@@ -256,9 +260,9 @@ fn parse_program_mod(ctx: &CrateContext) -> syn::ItemMod {
         })
         .collect::<Vec<_>>();
     if mods.len() != 1 {
-        panic!("Did not find program attribute");
+        return None;
     }
-    mods[0].clone()
+    Some(mods[0].clone())
 }
 
 fn parse_error_enum(ctx: &CrateContext) -> Option<syn::ItemEnum> {
